name = "arenas"
from .arena_core import Environment
from .simple2d import Simple2D
from .connected_rooms import ConnectedRooms
from .hafting_2008 import Hafting2008
from .sargolini_2006 import Sargolini2006, BasicSargolini2006
from .wernle_2018 import Wernle2018, MergingRoom
